import qalgo as qa
from qalgo import qda
import pysparq as sq

def test_import():
    sq.test_import()


def test_simple_program():
    sq.test_simple_program()


if __name__ == '__main__':
    test_import()
    test_simple_program()